/*
 * capchroot.c
 *
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License
 *      version 2 as published by the Free Software Foundation.
 *
 * Copyright (c) 2009 Thomas Bächler <thomas@archlinux.org>
 */

#ifndef CONFIGFILE
#error CONFIGFILE not defined
#endif

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <string.h>
#include <pwd.h>
#include <grp.h>
#include <sys/capability.h>
#include "configfile.h"

void usage(const char *progname) {
  fprintf(stderr, "Usage: %s [-d directory] path -- [command [args]]\n", progname);
}

int allowed(const char *path) {
  int ret = 0;
  struct pathcfg cfg;
  struct passwd *user;
  struct group *g;
  char **c, **gr;

  user = getpwuid(getuid());
  if(user == NULL) {
    perror("getpwuid");
  }

  /* Initialize the cfg structure */
  if(getconfig(path, &cfg)) {
    /* Check if the user was explicitly allowed */
    for(c = cfg.users; *c != NULL; ++c) {
      if((ret = !strcmp(*c, user->pw_name)))
        break;
    }

    if(!ret) {
      for(c = cfg.groups; *c != NULL; ++c) {
        g = getgrnam(*c);
        if(g == NULL)
          continue;
        /* First check if the group name is the primary group */
        if((ret = g->gr_gid == user->pw_gid))
          break;
        /* Iterate over all group members to see if the user is in the group */
        for(gr = g->gr_mem; *gr != NULL; gr++) {
          if((ret = !strcmp(*gr, user->pw_name))) {
            break;
          }
        }
      }
    }
    /* Free the components of cfg */
    getconfig_free(&cfg);
  }
  return ret;
}

int main(int argc, char *argv[]) {
  struct stat st;
  cap_t mycap;
  cap_flag_value_t r1, r2;
  int opt;
  char *dir = "/", *path;

  /* Check if the cap_sys_chroot capability is set, abort if it isn't */
  mycap = cap_get_proc();
  if(mycap == NULL) {
    perror("cap_get_proc");
    return -1;
  }
  cap_get_flag(mycap, CAP_SYS_CHROOT, CAP_PERMITTED, &r1);
  cap_get_flag(mycap, CAP_SYS_CHROOT, CAP_EFFECTIVE, &r2);
  if(r1 != CAP_SET || r2 != CAP_SET) {
    fprintf(stderr, "The cap_sys_chroot capability is missing, this program won't work.\n"
                    "Run 'setcap cap_sys_chroot=ep %s' to set the capability.\n", argv[0]);
    /*return -1;*/
  }
  cap_free(mycap);

  while((opt = getopt(argc, argv, "d:")) != -1) {
    switch(opt) {
      case 'd':
        dir = optarg;
        break;
      default:
        usage(argv[0]);
        return -1;
    }
  }

  /* Require at least one argument */
  if(argc - optind < 1) {
    usage(argv[0]);
    return -1;
  }

  /* Check if the configuration file exists, is owned by root and only writable by root */
  if(stat(CONFIGFILE, &st) != 0) {
    perror("stat: " CONFIGFILE);
    return -1;
  }
  if(st.st_uid != 0) {
    fprintf(stderr, "Configuration file " CONFIGFILE " is not owned by root.\n");
    return -1;
  }
  if(st.st_mode & (S_IWGRP | S_IWOTH)) {
    fprintf(stderr, "Configuration file " CONFIGFILE " must only be writable by root.\n");
    return -1;
  }

  /* Get the real path */
  path = realpath(argv[optind], NULL);
  if(path == NULL) {
    perror(argv[optind]);
    return -1;
  }

  /* Check if the specified directory is allowed by the configuration file */
  if(!allowed(path)) {
    fprintf(stderr, "Changing root to %s is not allowed\n", path);
    free(path);
    return -1;
  }

  /* Change the root directory */
  if(chroot(path) == -1) {
    perror("chroot");
    free(path);
    return -1;
  }
  free(path);

  /* Change the directory so we are actually inside the new root */
  if(chdir(dir) == -1) {
    perror("chdir");
    return -1;
  }

  if(argc-optind < 2) {
    /* If no further arguments are specified, execute an interactive shell */
    char *shell = getenv("SHELL");
    if(shell == NULL)
      shell = (char*)"/bin/sh";
    argv[0] = shell;
    argv[1] = (char*)"-i";
    argv[2] = NULL;
  }
  else {
    /* Execute the commands specified on the commandline */
    argv += optind + 1;
  }

  execvp(argv[0], argv);
  /* If execvp returns, something went wrong, fail gracefully */
  perror("execvp");
  return -1;
}
